#include <algorithm>
#include <bitset>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <cstring>

using namespace std;

struct Edge { int a, b; };

const int MAXN = 10000;

Edge edges[MAXN * MAXN];
bool outputed_edges[MAXN][MAXN];
int vertex_permutation[MAXN];
vector<int> neighbors[MAXN]; // almost_no_four_cycle
bitset<MAXN> neighbors_bitset[MAXN]; // almost_no_four_cycle, triangles_with_common_triangle
Edge skipped_edges[MAXN * MAXN]; // almost_no_four_cycle

unsigned get_random_uint16() {
  return rand() & ((1 << 16) - 1);
}

unsigned get_random_uint32() {
  return (get_random_uint16() << 16) | get_random_uint16();
}

void decode (int v, int p, int vd[])
{
  if (v == 0)
    {
      vd[0] = vd[1] = 0;
      vd[2] = 1;
      return;
    }
  v--;
  if (v < p)
    {
      vd[0] = 0;
      vd[1] = 1;
      vd[2] = v;
      return;
    }
  v -= p;
  vd[0] = 1;
  vd[1] = v / p;
  vd[2] = v % p;
}

bool incident (int v, int l, int p)
{
  int vd[3], ld[3];

  decode (v, p, vd);
  decode (l, p, ld);

  return (vd[0] * ld[0] + vd[1] * ld[1] + vd[2] * ld[2]) % p == 0;
}

int main(int argc, char **argv) {
  if (argc != 6) {
    fprintf(
      stderr,
      "Usage: ./gen-input <seed> <graph_type> <n> <max_m> <final_add>\n"
    );
    return 1;
  }
  int seed = atoi(argv[1]);
  const char *graph_type = argv[2];
  int n = atoi(argv[3]);
  int max_m = atoi(argv[4]);
  int final_add = atoi(argv[5]);
  srand(seed);
  assert(n >= 1 && n <= MAXN);
  int m = 0;
  if (strcmp(graph_type, "random") == 0) {
    assert(final_add == -1);
    for (int i = 0; i < n; i++) {
      for (int j = i + 1; j < n; j++) {
        edges[m++] = (Edge) { i, j };
      }
    }
    for (int i = m - 1; i > 0; i--) {
      int j = get_random_uint32() % i;
      swap(edges[i], edges[j]);
    }
    assert(max_m >= 0 && max_m <= m);
    m = max_m;
  } else if (strcmp(graph_type, "almost_no_four_cycle") == 0) {
    assert(max_m >= final_add);
    int m_all = 0;
    for (int i = 0; i < n; i++) {
      for (int j = i + 1; j < n; j++) {
        edges[m_all++] = (Edge) { i, j };
      }
    }
    for (int i = m_all - 1; i > 0; i--) {
      int j = get_random_uint32() % i;
      swap(edges[i], edges[j]);
    }
    int skipped = 0;
    for (int i = 0; i < m_all; i++) {
      Edge e = edges[i];
      if (m + final_add >= max_m) {
        if (skipped < final_add) {
          skipped_edges[skipped++] = e;
          continue;
        }
        break;
      }
      int h1 = e.a, h2 = e.b;
      for (int h3 : neighbors[h2]) {
        assert(h3 != h1);
        bitset<MAXN> h4 = neighbors_bitset[h3] & neighbors_bitset[h1];
        h4.reset(h2);
        if (h4.count() >= 1) {
          skipped_edges[skipped++] = e;
          goto next_edge;
        }
      }
      neighbors[e.a].push_back(e.b);
      neighbors[e.b].push_back(e.a);
      neighbors_bitset[e.a].set(e.b);
      neighbors_bitset[e.b].set(e.a);
      edges[m++] = e;
      next_edge:;
    }
    assert(final_add >= 0 && final_add <= skipped);
    for (int i = 0; i < final_add; i++) {
      edges[m++] = skipped_edges[i];
    }
    assert(m <= max_m);
  } else if (strcmp(graph_type, "triangles_with_common_vertex") == 0) {
    assert(max_m == -1);
    assert(final_add >= 0);
    int prev = -1;
    for (int i = 1; i < n; i++) {
      edges[m++] = (Edge) { 0, i };
      if (prev == -1) {
        prev = i;
      } else {
        edges[m++] = (Edge) { prev, i };
        prev = -1;
      }
    }
    for (int i = 0; i < m; i++) {
      Edge e = edges[i];
      neighbors_bitset[e.a].set(e.b);
      neighbors_bitset[e.b].set(e.a);
    }
    for (int i = 0; i < final_add; i++) {
      int a, b;
      do {
        a = get_random_uint32() % n;
        b = get_random_uint32() % n;
      } while (a == b || neighbors_bitset[a][b]);
      edges[m++] = (Edge) { a, b };
      neighbors_bitset[a].set(b);
      neighbors_bitset[b].set(a);
    }
  } else if (strcmp(graph_type, "triangles_with_common_triangle") == 0) {
    assert(max_m == -1);
    assert(final_add >= 0);
    edges[m++] = (Edge) { 0, 1 };
    edges[m++] = (Edge) { 0, 2 };
    edges[m++] = (Edge) { 1, 2 };
    int t = 0, prev = -1;
    for (int i = 3; i < n; i++) {
      edges[m++] = (Edge) { t, i };
      if (prev == -1) {
        prev = i;
      } else {
        edges[m++] = (Edge) { prev, i };
        prev = -1;
        t = (t + 1) % 3;
      }
    }
    for (int i = 0; i < m; i++) {
      Edge e = edges[i];
      neighbors_bitset[e.a].set(e.b);
      neighbors_bitset[e.b].set(e.a);
    }
    for (int i = 0; i < final_add; i++) {
      int a, b;
      do {
        a = get_random_uint32() % n;
        b = get_random_uint32() % n;
      } while (a == b || neighbors_bitset[a][b]);
      edges[m++] = (Edge) { a, b };
      neighbors_bitset[a].set(b);
      neighbors_bitset[b].set(a);
    }
  } else if (strcmp(graph_type, "proj_incid") == 0) {
      /* The incidence graph of the projective plane with n^2+n+1 points, where n is a prime.
         Both points and lines are equivalence classes of non-zero triples (a,b,c) of
	 elements of Z_n, where two triples are equivalent if they are linearly dependent.
	 Thus, we can select (1,b,c), (0,1,c), and (0,0,1) as representatives.  Point is
	 incident with a line iff their cross-product is zero.  */
      int p = n;
      n = p * p + p + 1;
      for (int v = 0; v < n; v++)
	for (int l = 0; l < n; l++)
	  if (incident (v, l, p))
	    edges[m++] = (Edge) {v, n + l};
      n *= 2;
      if (final_add)
	{
	  edges[m++] = (Edge) {n-1, n};
	  edges[m++] = (Edge) {n-1, n+1};
	  edges[m++] = (Edge) {n, n+2};
	  edges[m++] = (Edge) {n+1, n+2};
	  n += 3;
	}
  } else {
    assert(false);
  }
  for (int i = 0; i < n; i++) {
    vertex_permutation[i] = i;
  }
  for (int i = n - 1; i > 0; i--) {
    int j = get_random_uint32() % i;
    swap(vertex_permutation[i], vertex_permutation[j]);
  }
  for (int i = m - 1; i > 0; i--) {
    int j = get_random_uint32() % i;
    swap(edges[i], edges[j]);
  }
  printf("%d %d\n", n, m);
  for (int i = 0; i < m; i++) {
    Edge e = edges[i];
    int a = vertex_permutation[e.a];
    int b = vertex_permutation[e.b];
    if (a > b) swap(a, b);
    assert(0 <= a && a < b && b < n);
    assert(!outputed_edges[a][b]);
    outputed_edges[a][b] = true;
    printf("%d %d\n", a + 1, b + 1);
  }
  return 0;
}
